import re

from nalej_platformer.exceptions import OrganizationPlanException
from nalej_platformer.utils import print_debug, spinner_ko_message, spinner_ok_message, get_indentation, get_spinner


class OrganizationUserPlan:
    def __init__(self, name, email, password):
        self.name = name
        self.email = email
        self.password = password

    def info(self, indentation=3):
        indent = get_indentation(indentation)
        info_template = (
            f"{indent}Name: {self.name}\n"
            f"{indent}Email: {self.email}"
        )
        return info_template

    @classmethod
    def read_plan(cls, plan_data, user_type):
        print_debug(
            f"Reading {user_type} user block. Raw data: {plan_data}")
        if plan_data is None:
            cap_user_type = user_type.capitalize()
            raise OrganizationPlanException(f"{cap_user_type} user block is required")

        with get_spinner(f"Checking {user_type} user block...") as spinner:
            name = cls._validate_name(plan_data.get("name", None), user_type, spinner)
            email = cls._validate_email(plan_data.get("email", None), user_type, spinner)
            password = cls._validate_password(plan_data.get("password", None), user_type, spinner)

        return cls(name, email, password)

    @classmethod
    def _validate_name(cls, name, user_type, spinner):
        cap_user_type = user_type.capitalize()
        if name in ["", None]:
            spinner_ko_message(spinner, f"{cap_user_type} user name is empty")
            raise OrganizationPlanException(f"{cap_user_type} user name is empty")

        spinner_ok_message(spinner, f"{cap_user_type} user name is valid")
        return name

    @classmethod
    def _validate_email(cls, email, user_type, spinner):
        cap_user_type = user_type.capitalize()
        if email in ["", None]:
            spinner_ko_message(spinner, f"{cap_user_type} user email is empty")
            raise OrganizationPlanException(f"{cap_user_type} user email is empty")

        email_regex = r"(^[a-zA-Z0-9_.+-]+@[a-zA-Z0-9-]+\.[a-zA-Z0-9-.]+$)"
        if re.match(email_regex, email) is None:
            spinner_ko_message(spinner, f"{cap_user_type} user email is not valid")
            raise OrganizationPlanException(f"{cap_user_type} user email is not valid")

        spinner_ok_message(spinner, f"{cap_user_type} user email is valid")
        return email

    @classmethod
    def _validate_password(cls, password, user_type, spinner):
        cap_user_type = user_type.capitalize()
        if password in ["", None]:
            spinner_ko_message(spinner, f"{cap_user_type} user password is empty")
            raise OrganizationPlanException(f"{cap_user_type} user password is empty")
        if len(password) < 6:
            spinner_ko_message(spinner, f"{cap_user_type} user password doesn't match the password policy")
            raise OrganizationPlanException(f"{cap_user_type} user password needs at least 6 characters")

        spinner_ok_message(spinner, f"{cap_user_type} user password is valid")
        return password

    def dict_for_yaml(self):
        return {
            "name": self.name,
            "email": self.email,
            "password": self.password
        }



from datetime import datetime, timedelta
from tempfile import gettempdir
import subprocess
import json

import click

from nalej_platformer.utils import spinner_ok_message, spinner_ko_message, print_debug
from nalej_platformer.exceptions import ProvisionerProviderException


def _get_path():
    ctx = click.get_current_context()
    config = ctx.obj['CONFIGURATION']
    return config.provisioner_path


def _get_resources_path():
    ctx = click.get_current_context()
    config = ctx.obj['CONFIGURATION']
    return config.provisioner_resources


def provision_management_cluster(
        sp_path, vm_size, region, name, resource_group, dns_zone, kubernetes_version,
        num_nodes, environment, spinner):

    start_time = datetime.now()
    cmd = (
        f'{_get_path()} provision --debug --azureCredentialsPath={sp_path} --platform AZURE --nodeType {vm_size} '
        f'--zone "{region}" --name {name} --resourceGroup {resource_group} --tempPath={gettempdir()} '
        f'--resourcesPath={_get_resources_path()} --dnsZoneName={dns_zone} --kubernetesVersion={kubernetes_version} '
        f'--numNodes={num_nodes}'
    )
    if environment == "PRODUCTION":
        cmd += ' --isProduction'

    provision_mngt = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True, shell=True)

    result_data = dict()
    while True:
        if provision_mngt.poll() is not None:
            break

        stderr_line = provision_mngt.stderr.readline()
        debug_line = stderr_line.strip('\n')
        if debug_line != "":
            print_debug(f"Provision output: {debug_line}")
        elapsed_time = datetime.now() - start_time
        elapsed_time = elapsed_time - timedelta(microseconds=elapsed_time.microseconds)
        spinner.text = f'Provisioning management cluster...[{elapsed_time}]'

        try:
            line_json = json.loads(stderr_line)
        except json.JSONDecodeError:
            pass
        else:
            line_type = line_json.get("type", None)
            line_progress = line_json.get("progress", None)
            if line_type == "Provision" and line_progress == "Finished":
                result_data = line_json

    provision_mngt.stdout.close()
    provision_mngt.stderr.close()
    exit_code = provision_mngt.poll()
    if exit_code == 0:
        spinner_ok_message(spinner, "Management cluster provisioned successfully")
    else:
        spinner_ko_message(spinner, "Error provisioning management cluster")
        raise ProvisionerProviderException("Error provisioning management cluster")

    return result_data



import click
import os

BASE_PATH = click.get_app_dir('nalej-platformer', force_posix=True)
CONFIG_PATH = os.path.join(BASE_PATH, "config.yaml")
SP_PATH = os.path.join(BASE_PATH, ".azure-sp.json")
KC_PATH = os.path.join(BASE_PATH, "kubeconfigs")



import sys

import click
import crayons
from vistir.path import normalize_path

from nalej_platformer.exceptions import PlanException, ConfigException, AzureProviderException
from nalej_platformer.config import defaults as cfg_defaults
from nalej_platformer.cli.config import load_configuration
from nalej_platformer.plan import Plan
from nalej_platformer.provider.azure import login


@click.group()
@click.option(
    '--config', default=cfg_defaults.CONFIG_PATH, show_default=True, type=click.Path(file_okay=True, dir_okay=False),
    help='Path to the configuration file')
@click.option('--resources', default="", type=click.Path(file_okay=False, dir_okay=True), help='Path to resources')
@click.option('--debug', default=False, is_flag=True, help='Toggle debug mode')
@click.pass_context
def run(ctx, config, resources, debug):
    """nalej-platformer is a helper utility to simplify the life cycle of a Nalej Platform from a CLI"""

    ctx.ensure_object(dict)
    ctx.obj['CONFIG_PATH'] = normalize_path(config)
    ctx.obj['RESOURCES_PATH'] = normalize_path(resources)
    ctx.obj['DEBUG'] = debug
    try:
        load_configuration(ctx)
        login()
        click.echo("Startup OK\n")
    except (ConfigException, AzureProviderException) as exc:
        click.echo(crayons.red(exc, bold=True))
        sys.exit(1)


@run.command()
@click.argument("plan", type=click.Path(exists=True))
@click.pass_context
def info(ctx, plan):
    """Evaluates a plan and outputs a human readable version of it"""

    ctx.obj['PLAN_PATH'] = normalize_path(plan)
    try:
        plan = Plan.read_plan(ctx.obj['PLAN_PATH'])
        click.echo("############################")
        click.echo("##   PLAN DETAILED VIEW   ##")
        click.echo("############################\n")
        click.echo(plan.info())
    except PlanException as pe:
        click.echo(crayons.red(pe, bold=True))
        sys.exit(1)


@run.command()
@click.argument("plan", type=click.Path(exists=True))
@click.option(
    '--force', '-f', is_flag=True, default=False, help="Force the apply of the plan without user confirmation")
@click.pass_context
def apply(ctx, plan, force):
    """Applies the contents of a plan"""

    ctx.obj['PLAN_PATH'] = normalize_path(plan)
    try:
        plan = Plan.read_plan(ctx.obj['PLAN_PATH'])

        if force or click.confirm('Do you want to continue?', default=True, abort=True):
            plan.apply()
    except PlanException as pe:
        click.echo(crayons.red(pe, bold=True))
        sys.exit(1)



import subprocess
import json
from datetime import datetime, timedelta
import time

import click

from nalej_platformer.utils import spinner_ok_message, spinner_ko_message, print_debug
from nalej_platformer.exceptions import KubernetesProviderException


def _get_path():
    ctx = click.get_current_context()
    config = ctx.obj['CONFIGURATION']
    return config.kubectl_path


def validate_cluster_installation(kubeconfig, spinner, oneshot=False):
    start_time = datetime.now()
    results = []
    node_count = get_node_count(kubeconfig)
    for i in range(0, 36):
        elapsed_time = datetime.now() - start_time
        elapsed_time = elapsed_time - timedelta(microseconds=elapsed_time.microseconds)
        spinner.text = f'Validating cluster installation...[{elapsed_time}]'

        results = [
            _validate_replicable_resource(kubeconfig, "deployments", spinner),
            _validate_replicable_resource(kubeconfig, "statefulsets", spinner),
            _validate_daemonsets(kubeconfig, node_count, spinner),
            _validate_jobs(kubeconfig, spinner)
        ]

        if all(results):
            spinner_ok_message(spinner, "Cluster installation is valid")
            break

        if oneshot:
            break

        print_debug("Platform is not ready yet, waiting 30 seconds...")
        time.sleep(30)

    return all(results)


def get_node_count(kubeconfig):
    try:
        print_debug(f'Getting node count: {kubeconfig}')
        cmd = f"{_get_path()} --kubeconfig {kubeconfig} -nnalej get nodes -o json"
        get_nodes = subprocess.run(cmd, shell=True, check=True, capture_output=True, text=True)
    except subprocess.CalledProcessError as cpe:
        print_debug(f'Error getting node count from cluster. Output: {cpe.stdout}. Error: {cpe.stderr}.')
        raise KubernetesProviderException(f"Error getting node count from cluster")
    else:
        try:
            output = json.loads(get_nodes.stdout)
        except json.JSONDecodeError as jexc:
            print_debug(f'Malformed response from Kubernetes. Output: {jexc}.')
            raise KubernetesProviderException("Malformed response from Kubernetes")
        else:
            return len(output.get('items', []))


def _validate_replicable_resource(kubeconfig, resource_kind, spinner):
    try:
        print_debug(f'Validating all {resource_kind} are ready: {kubeconfig}')
        cmd = f"{_get_path()} --kubeconfig {kubeconfig} -nnalej get {resource_kind} -o json"
        get_resource = subprocess.run(cmd, shell=True, check=True, capture_output=True, text=True)
    except subprocess.CalledProcessError as cpe:
        print_debug(f'Error getting {resource_kind} from cluster. Output: {cpe.stdout}. Error: {cpe.stderr}.')
        spinner_ko_message(spinner, f"Error getting {resource_kind} from cluster")
        raise KubernetesProviderException(f"Error getting {resource_kind} from cluster")
    else:
        try:
            output = json.loads(get_resource.stdout)
        except json.JSONDecodeError as jexc:
            print_debug(f'Malformed response from Kubernetes. Output: {jexc}.')
            spinner_ko_message(spinner, "Malformed response from Kubernetes")
            raise KubernetesProviderException("Malformed response from Kubernetes")
        else:
            for resource in output['items']:
                replicas = int(resource['spec']['replicas'])
                ready_output = resource['status'].get('readyReplicas', None)
                if ready_output in ["", None]:
                    ready_output = 0
                ready_replicas = int(ready_output)
                print_debug((
                    f"{resource_kind.capitalize()} {resource['metadata']['name']}, "
                    f"found {replicas} replicas which {ready_replicas} are ready")
                )
                if replicas != ready_replicas:
                    return False
            print_debug(f"All {resource_kind} are ready")
            return True


def _validate_daemonsets(kubeconfig, node_count, spinner):
    try:
        print_debug(f'Validating all daemonsets are ready: {kubeconfig}')
        cmd = f"{_get_path()} --kubeconfig {kubeconfig} -nnalej get daemonsets -o json"
        get_daemonsets = subprocess.run(cmd, shell=True, check=True, capture_output=True, text=True)
    except subprocess.CalledProcessError as cpe:
        print_debug(f'Error getting daemonsets from cluster. Output: {cpe.stdout}. Error: {cpe.stderr}.')
        spinner_ko_message(spinner, f"Error getting daemonsets from cluster")
        raise KubernetesProviderException(f"Error getting daemonsets from cluster")
    else:
        try:
            output = json.loads(get_daemonsets.stdout)
        except json.JSONDecodeError as jexc:
            print_debug(f'Malformed response from Kubernetes. Output: {jexc}.')
            spinner_ko_message(spinner, "Malformed response from Kubernetes")
            raise KubernetesProviderException("Malformed response from Kubernetes")
        else:
            for daemonset in output['items']:
                ready_output = daemonset['status'].get('numberReady', None)
                if ready_output in ["", None]:
                    ready_output = 0
                ready_output = int(ready_output)
                print_debug((
                    f"Daemonset {daemonset['metadata']['name']}, "
                    f"expected {node_count} replicas which {ready_output} are ready")
                )
                if node_count != ready_output:
                    return False
            print_debug(f"All daemonsets are ready")
            return True


def _validate_jobs(kubeconfig, spinner):
    try:
        print_debug(f'Validating all jobs are complete: {kubeconfig}')
        cmd = f"{_get_path()} --kubeconfig {kubeconfig} -nnalej get jobs -o json"
        get_jobs = subprocess.run(cmd, shell=True, check=True, capture_output=True, text=True)
    except subprocess.CalledProcessError as cpe:
        print_debug(f'Error getting jobs from cluster. Output: {cpe.stdout}. Error: {cpe.stderr}.')
        spinner_ko_message(spinner, f"Error getting jobs from cluster")
        raise KubernetesProviderException(f"Error getting jobs from cluster")
    else:
        try:
            output = json.loads(get_jobs.stdout)
        except json.JSONDecodeError as jexc:
            print_debug(f'Malformed response from Kubernetes. Output: {jexc}.')
            spinner_ko_message(spinner, "Malformed response from Kubernetes")
            raise KubernetesProviderException("Malformed response from Kubernetes")
        else:
            for job in output['items']:
                succeeded = job['status'].get('succeeded', 0)
                if succeeded in ["", None, 0] or int(succeeded) <= 0:
                    print_debug(f"Job {job['metadata']['name']} has no succeeded jobs")
                    return False
            print_debug("Every job has succeeded")
            return True



from nalej_platformer.exceptions import HostnamesPlanException
from nalej_platformer.utils import print_debug, get_indentation


class HostnamesPlan:
    def __init__(self, web, api, login, signup):
        self.web = web
        self.api = api
        self.login = login
        self.signup = signup

    def info(self, indentation=2):
        indent = get_indentation(indentation)
        info_template = (
            f"{indent}Web: {self.web}\n"
            f"{indent}API: {self.api}\n"
            f"{indent}Login: {self.login}\n"
            f"{indent}SignUp: {self.signup}\n"
        )

        return info_template

    @classmethod
    def read_plan(cls, plan_data):
        print_debug(f"Reading hostnames plan block. Raw data: {plan_data}")
        if plan_data is None:
            return None

        web = plan_data.get("web", None)
        api = plan_data.get("api", None)
        login = plan_data.get("login", None)
        signup = plan_data.get("signup", None)

        if web is None:
            raise HostnamesPlanException("Web hostname not found in plan")
        if api is None:
            raise HostnamesPlanException("API hostname not found in plan")
        if login is None:
            raise HostnamesPlanException("Login hostname not found in plan")
        if signup is None:
            raise HostnamesPlanException("SignUp hostname not found in plan")

        return cls(web, api, login, signup)

    def dict_for_yaml(self):
        return {
            "web": self.web,
            "api": self.api,
            "login": self.login,
            "signup": self.signup
        }



import subprocess
import time
import json
from datetime import datetime, timedelta

import click

from nalej_platformer.utils import spinner_ok_message, spinner_ko_message, print_debug, parse_json_output
from nalej_platformer.exceptions import SignupProviderException


def _get_path():
    ctx = click.get_current_context()
    config = ctx.obj['CONFIGURATION']
    return config.signup_path


def _get_ca_path(environment):
    ctx = click.get_current_context()
    config = ctx.obj['CONFIGURATION']
    return config.get_ca_path(environment)


def wait_for_signup(signup_address, environment, spinner):
    start_time = datetime.now()
    print_debug("Waiting for signup to be ready")
    for i in range(0, 36):
        elapsed_time = datetime.now() - start_time
        elapsed_time = elapsed_time - timedelta(microseconds=elapsed_time.microseconds)
        spinner.text = f'Waiting for signup to be ready...[{elapsed_time}]'
        try:
            cmd = f'{_get_path()} list --signupAddress={signup_address}:443 --caPath={_get_ca_path(environment)}'
            subprocess.run(cmd, shell=True, check=True, capture_output=True, text=True)
        except subprocess.CalledProcessError:
            print_debug("Signup is not ready yet, waiting 5 seconds")
            time.sleep(5)
            continue
        else:
            print_debug("Signup is ready")
            spinner_ok_message(spinner, "Signup component is ready")
            return True
    print_debug("Timeout waiting for signup to be ready")
    spinner_ko_message(spinner, "Timeout waiting for signup to be ready")
    raise SignupProviderException("Timeout waiting for signup to be ready")


def create_organization(signup_address, organization, environment, spinner):
    print_debug(f'Creating organization {organization.name}')
    owner = organization.owner
    nalej_admin = organization.nalej_admin
    try:
        cmd = (
            f'{_get_path()} signup --signupAddress={signup_address}:443 --orgName={organization.name} '
            f'--ownerEmail={owner.email} --ownerName={owner.email} --ownerPassword={owner.password} '
            f'--nalejAdminEmail={nalej_admin.email} --nalejAdminName={nalej_admin.email} '
            f'--nalejAdminPassword={nalej_admin.password} --caPath={_get_ca_path(environment)}'
        )
        create_org = subprocess.run(cmd, shell=True, check=True, capture_output=True, text=True)
    except subprocess.CalledProcessError as cpe:
        print_debug(f'Error creating organization {organization.name}. Output: {cpe.stdout}. Error: {cpe.stderr}.')
        spinner_ko_message(spinner, f"Error creating organization {organization.name}")
        raise SignupProviderException(f"Error creating organization {organization.name}")
    else:
        try:
            parsed_json = parse_json_output(create_org.stderr)
        except json.JSONDecodeError:
            print_debug(f'Malformed response from signup.')
            spinner_ko_message(spinner, f"Malformed response from organization")
            raise SignupProviderException(f"Malformed response from signup")
        else:
            if len(parsed_json) < 1:
                print_debug(f'Malformed response from signup. Output: {parsed_json}.')
                spinner_ko_message(spinner, f"Malformed response from organization")
                raise SignupProviderException(f"Malformed response from signup")
            signup_data = parsed_json[-1]
            organization_id = signup_data.get("organizationID", None)
            if organization_id is None:
                print_debug(f'Organization ID not found in response. Output: {parsed_json}.')
                spinner_ko_message(spinner, f"Organization ID not found in response")
                raise SignupProviderException(f"Organization ID not found in response")
            else:
                print_debug(f'Organization {organization.name} created with {organization_id}')
                spinner_ok_message(spinner, f"Organization {organization.name} created successfully")
                return organization_id


def organization_exists(signup_address, organization, environment):
    if organization.organization_id is None:
        return False

    print_debug(f'Checking if organization {organization.name} exists')
    try:
        cmd = (
            f'{_get_path()} info  --signupAddress={signup_address}:443 '
            f'--organizationID={organization.organization_id} --caPath={_get_ca_path(environment)}'
        )
        subprocess.run(cmd, shell=True, check=True, capture_output=True)
    except subprocess.CalledProcessError:
        return False
    else:
        return True



class ConfigException(Exception):
    """Exception raised when the configuration file is not valid"""

    def __init__(self, message):
        self.message = message


class PlanException(Exception):
    """Exception raised if something in the plan is not valid"""

    def __init__(self, message):
        self.message = message


class ManagementPlanException(PlanException):
    """Exception raised if something in the management plan is not valid"""

    def __init__(self, message):
        self.message = message


class ApplicationPlanException(PlanException):
    """Exception raised if something in the application plan is not valid"""

    def __init__(self, message):
        self.message = message


class AzureConfigPlanException(PlanException):
    """Exception raised if something in the azure config block is not valid"""

    def __init__(self, message):
        self.message = message


class KubernetesConfigPlanException(PlanException):
    """Exception raised if something in the kubernetes config block is not valid"""

    def __init__(self, message):
        self.message = message


class OrganizationPlanException(PlanException):
    """Exception raised if something in the organization config block is not valid"""

    def __init__(self, message):
        self.message = message


class HostnamesPlanException(PlanException):
    """Exception raised if something in the hostnames block is not valid"""

    def __init__(self, message):
        self.message = message


class IPAddressesPlanException(PlanException):
    """Exception raised if something in the IP Addresses block is not valid"""

    def __init__(self, message):
        self.message = message


class ProviderException(Exception):
    """Base exception for provider related exceptions"""

    def __init__(self, message):
        self.message = message


class AzureProviderException(ProviderException):
    """Exception using Azure provider calls"""

    def __init__(self, message):
        self.message = message


class KubernetesProviderException(ProviderException):
    """Exception using Kubernetes provider calls"""

    def __init__(self, message):
        self.message = message


class InstallerProviderException(ProviderException):
    """Exception using Installer provider calls"""

    def __init__(self, message):
        self.message = message


class ProvisionerProviderException(ProviderException):
    """Exception using Provisioner provider calls"""

    def __init__(self, message):
        self.message = message


class SignupProviderException(ProviderException):
    """Exception using Signup provider calls"""

    def __init__(self, message):
        self.message = message


class PublicApiProviderException(ProviderException):
    """Exception using Public API provider calls"""

    def __init__(self, message):
        self.message = message



import json
import io

import crayons
import click
from yaspin import yaspin


def print_debug(message):
    ctx = click.get_current_context()

    if ctx.obj['DEBUG']:
        debug_msg = f"({crayons.blue('DEBUG', bold=True)}) {message}"
        spinner = ctx.obj.get('SPINNER', None)
        if spinner is not None:
            spinner.write(debug_msg)
        else:
            click.echo(debug_msg)


def rreplace(s, old, new, occurrence):
    li = s.rsplit(old, occurrence)
    return new.join(li)


def spinner_ok_message(spinner, message):
    spinner.text = message
    spinner.ok("✅")


def spinner_ko_message(spinner, message):
    spinner.text = message
    spinner.fail("💥")


def get_spinner(text):
    new_spinner = yaspin(color="blue", text=text)
    ctx = click.get_current_context()
    ctx.obj['SPINNER'] = new_spinner
    return new_spinner


def copy_kubeconfig(source, destination):
    with open(source, "r") as source_file:
        with open(destination, "w") as destination_file:
            destination_file.write(source_file.read())


def get_indentation(level):
    return " " * 4 * level


def parse_json_output(contents):
    if isinstance(contents, str):
        contents = io.StringIO(contents)
    contents.seek(0)
    parsed_file = [line.rstrip('\n').lstrip(" ") for line in contents.readlines()]
    fixed_content = "".join(parsed_file).replace("}{", "},{")
    return json.loads(f'[{fixed_content}]')



import os
import platform
from shutil import which

import yaml
import vistir

from nalej_platformer.utils import print_debug, spinner_ok_message, spinner_ko_message, get_spinner
from nalej_platformer.exceptions import ConfigException

HUMAN_READABLE = {
    "resources": "Platform resources",
    "azure": "Azure CLI",
    "kubectl": "Kubernetes CLI"
}

PARAMETER_TRANSFORM = {
    "resources": "resources_path",
    "azure": "az_path",
    "kubectl": "kubectl_path"
}

ENVIRONMENT_CA = {
    "DEVELOPMENT": "letsencrypt_staging.pem",
    "STAGING": "letsencrypt_staging.pem",
    "PRODUCTION": "letsencrypt_prod.pem"
}


class Configuration:
    def __init__(
            self, resources_path=None, az_path=None, kubectl_path=None):
        self.resources_path = resources_path
        self.az_path = az_path
        self.kubectl_path = kubectl_path

    def validate(self):
        for name, arg in PARAMETER_TRANSFORM.items():
            attr = getattr(self, arg, None)
            if attr is None or not os.path.exists(attr):
                error = (
                    f"Can't find or open {HUMAN_READABLE[name]}. "
                    "Please ensure that is in your PATH."
                )
                raise ConfigException(error)

    def is_valid(self):
        return all(
            [
                self.az_path is not None,
                self.kubectl_path is not None,
                self.resources_path is not None
            ]
        )

    def find_missing_binaries(self):
        if not self.is_valid():
            if self.az_path is None:
                self.az_path = which("az")
            if self.kubectl_path is None:
                self.kubectl_path = which("kubectl")

    @classmethod
    def load_from_file(cls, resources_path=None):
        """Loads the configuration file and creates the configuration instance"""

        configuration = {}
        # print_debug(f'Loading configuration file: {config_path}')
        # with get_spinner("Loading configuration file...") as spinner:
        #     with open(config_path, "r") as config_file:
        #         try:
        #             config_dict = yaml.load(config_file, Loader=yaml.FullLoader)
        #             paths = config_dict.get('paths', None)
        #             if paths is None:
        #                 paths = {}

        #             for binary, path in paths.items():
        #                 spinner.text = 'Reading {HUMAN_READABLE[binary]} path'
        #                 configuration[PARAMETER_TRANSFORM[binary]] = vistir.path.normalize_path(path)

        #             spinner_ok_message(spinner, "Configuration file loaded")
        #         except:
        #             spinner_ko_message(spinner, "Error loading configuration file")
        if resources_path not in ['', None]:
            configuration['resources_path'] = resources_path
        else:
            raise ConfigException("Resources path is not defined.")

        return cls(**configuration)

    def get_ca_path(self, environment):
        return os.path.join(self.provisioner_resources, "ca", ENVIRONMENT_CA[environment])

    def _get_os_binaries(self):
        os_name = platform.system().lower()
        return os.path.join(self.resources_path, "binaries", os_name)

    def _get_assets(self):
        return os.path.join(self.resources_path, "assets")

    @property
    def provisioner_resources(self):
        return os.path.join(self._get_assets(), "provisioner")

    @property
    def installer_resources(self):
        return os.path.join(self._get_assets(), "installer")

    @property
    def installer_binaries(self):
        return self._get_os_binaries()

    @property
    def provisioner_path(self):
        return os.path.join(self._get_os_binaries(), "provisioner-cli")

    @property
    def installer_path(self):
        return os.path.join(self._get_os_binaries(), "installer-cli")

    @property
    def public_api_path(self):
        return os.path.join(self._get_os_binaries(), "public-api-cli")

    @property
    def signup_path(self):
        return os.path.join(self._get_os_binaries(), "signup-cli")



import subprocess

import click

from nalej_platformer.utils import spinner_ok_message, spinner_ko_message, print_debug, parse_json_output
from nalej_platformer.exceptions import PublicApiProviderException


def _get_path():
    ctx = click.get_current_context()
    config = ctx.obj['CONFIGURATION']
    return config.public_api_path


def _get_ca_path(environment):
    ctx = click.get_current_context()
    config = ctx.obj['CONFIGURATION']
    return config.get_ca_path(environment)


def login(login_address, environment, user, spinner):
    print_debug(f"Logging into the platform with user {user.email}")
    try:
        cmd = (
            f'{_get_path()} login --loginAddress={login_address} --loginPort=443 --email={user.email} '
            f'--password={user.password} --cacert={_get_ca_path(environment)}'
        )
        subprocess.run(cmd, shell=True, check=True, capture_output=True, text=True)
    except subprocess.CalledProcessError as cpe:
        print_debug(
            f'Error logging to the platform with user {user.email}. Output: {cpe.stdout}. Error: {cpe.stderr}.')
        spinner_ko_message(spinner, f"Error logging to the platform with user {user.email}")
        raise PublicApiProviderException(f"Error logging to the platform with user {user.email}")
    else:
        print_debug("Logged in to the platform successfully")
        spinner_ok_message(spinner, "Logged in to the platform successfully")
        return True


def provision_and_install(
        sp_path, vm_size, region, name, resource_group, dns_zone, kubernetes_version,
        num_nodes, environment, api_address, organization_id, spinner):
    print_debug(f"Launching provision and install application cluster {name}")
    try:
        cmd = (
            f'{_get_path()} cluster provision-and-install --nalejAddress={api_address} --clusterName={name} '
            f'--azureDnsZoneName={dns_zone} --azureResourceGroup={resource_group} --numNodes={num_nodes} '
            f'--nodeType={vm_size} --zone="{region}" --kubernetesVersion={kubernetes_version} '
            f'--azureCredentialsPath={sp_path} --organizationId={organization_id} --targetPlatform=AZURE '
            f'--cacert={_get_ca_path(environment)} --clusterType=KUBERNETES'
        )
        if environment == "PRODUCTION":
            cmd += ' --isProductionCluster'

        pai_output = subprocess.run(cmd, shell=True, check=True, capture_output=True, text=True)
    except subprocess.CalledProcessError as cpe:
        print_debug((
            f'Error launching provision and install application cluster {name}. '
            f'Output: {cpe.stdout}. Error: {cpe.stderr}.'
        ))
        spinner_ko_message(spinner, f"Error launching provision and install application cluster {name}")
        raise PublicApiProviderException(f"Error launching provision and install application cluster {name}")
    else:
        print_debug("Launched provision and install process succesfully")
        json_output = parse_json_output(pai_output.stdout)[0]
        cluster_id = json_output.get("cluster_id", None)
        print_debug(f'Application cluster ID: {cluster_id}')
        spinner_ok_message(spinner, "Launched provision and install process succesfully")
        return json_output


def get_cluster_state(api_address, environment, cluster_id, spinner):
    print_debug(f'Validating that cluster {cluster_id} is installed')
    try:
        cmd = (
            f'{_get_path()} cluster info --nalejAddress={api_address} --cacert={_get_ca_path(environment)} '
            f'{cluster_id}'
        )

        cluster_info = subprocess.run(cmd, shell=True, check=True, capture_output=True, text=True)
    except subprocess.CalledProcessError as cpe:
        print_debug(f'Error validating cluster {cluster_id} is installed. Output: {cpe.stdout}. Error: {cpe.stderr}.')
        spinner_ko_message(spinner, f"Error validating cluster {cluster_id} is installed")
        raise PublicApiProviderException(f"Error validating cluster {cluster_id} is installed")
    else:
        cluster_json = parse_json_output(cluster_info.stdout)[0]
        state_name = cluster_json.get("state_name", None)
        if state_name is None:
            print_debug('Malformed response getting cluster info')
            spinner_ko_message(spinner, 'Malformed response getting cluster info')
            raise PublicApiProviderException('Malformed response getting cluster info')
        else:
            return state_name



import os

import vistir

from nalej_platformer.exceptions import KubernetesConfigPlanException, AzureProviderException
from nalej_platformer.utils import print_debug, spinner_ko_message, spinner_ok_message, copy_kubeconfig
from nalej_platformer.utils import get_indentation, get_spinner
from nalej_platformer.provider.azure import check_k8s_version
from nalej_platformer.config.defaults import KC_PATH


class KubernetesConfigPlan:
    def __init__(self, version, num_nodes, kubeconfig):
        self.version = version
        self.num_nodes = num_nodes
        self.kubeconfig = kubeconfig

    def info(self, indentation=2):
        indent = get_indentation(indentation)
        info_template = (
            f"{indent}Version: {self.version}\n"
            f"{indent}Number of nodes: {self.num_nodes}"
        )

        if self.kubeconfig is not None:
            info_template += f"\n{indent}Kubeconfig: {self.kubeconfig}."

        return info_template

    @classmethod
    def read_plan(cls, plan_data, region=None):
        print_debug(f"Reading kubernetes configuration block. Raw data: {plan_data}")
        if plan_data is None:
            raise KubernetesConfigPlanException("Kubernetes configuration block is required")

        try:
            version = cls._validate_version(plan_data.get("version", None), region)
            num_nodes = cls.validate_num_nodes(plan_data.get("nodes", None))
        except AzureProviderException as ape:
            raise KubernetesConfigPlanException(f"Error checking with Azure: {ape}")

        kubeconfig = cls._validate_kubeconfig(plan_data.get("kubeconfig", None))

        return cls(version, num_nodes, kubeconfig)

    @classmethod
    def _validate_version(cls, version, region=None):
        with get_spinner("Checking kubernetes version is valid...") as spinner:
            if version in ["", None]:
                spinner_ko_message(spinner, "Kubernetes version is empty")
                raise KubernetesConfigPlanException("Kubernetes version must not be empty")

            cls._check_minimum_k8s_version(version, spinner)

            if region is None:
                return version
            else:
                return check_k8s_version(version, region, spinner)

    @staticmethod
    def validate_num_nodes(num_nodes):
        with get_spinner("Checking number of nodes are valid...") as spinner:
            if num_nodes in [None, ""]:
                spinner_ko_message(spinner, "Number of nodes is empty")
                raise KubernetesConfigPlanException("Number of nodes must not be empty")

            if num_nodes < 3:
                spinner_ko_message(spinner, "Number of nodes are lower than 3")
                raise KubernetesConfigPlanException("The minimum number of nodes is 3")

            spinner_ok_message(spinner, "Number of nodes is valid")
            return num_nodes

    @staticmethod
    def _validate_kubeconfig(kubeconfig):
        if kubeconfig is not None:
            if os.path.exists(kubeconfig):
                return vistir.path.normalize_path(kubeconfig)
            else:
                raise KubernetesConfigPlanException("Kubeconfig file not found")

        return kubeconfig

    @staticmethod
    def _check_minimum_k8s_version(version, spinner):
        version_numbers = version.split(".")
        if len(version_numbers) < 2:
            spinner_ko_message(spinner, "Kubernetes version format is not valid")
            raise KubernetesConfigPlanException("Kubernetes version is not valid")

        try:
            major_version = int(version_numbers[0])
            minor_version = int(version_numbers[1])
        except ValueError:
            spinner_ko_message(spinner, "Kubernetes version format is not valid")
            raise KubernetesConfigPlanException("Kubernetes version is not valid")

        if major_version < 1 and minor_version < 11:
            spinner_ko_message(spinner, "Kubernetes version is not at least 1.11")
            raise KubernetesConfigPlanException("Kubernetes version is not valid")

    def _set_kubeconfig(self, source_path, cluster_name):
        if source_path is None:
            raise KubernetesConfigPlanException("Kubeconfig not present after provision")

        vistir.path.mkdir_p(KC_PATH)

        kc_dest_path = os.path.join(KC_PATH, f"{cluster_name}.yaml")
        copy_kubeconfig(source_path, kc_dest_path)

        self.kubeconfig = kc_dest_path
        try:
            os.remove(source_path)
        except FileNotFoundError:
            pass

    def dict_for_yaml(self):
        yaml_data = {
            "version": self.version,
            "nodes": self.num_nodes
        }

        if self.kubeconfig is not None:
            yaml_data["kubeconfig"] = self.kubeconfig

        return yaml_data



from nalej_platformer.utils import print_debug
from nalej_platformer.config import Configuration


def load_configuration(ctx):
    resources_path = ctx.obj['RESOURCES_PATH']
    config = Configuration.load_from_file(resources_path)

    config.find_missing_binaries()
    config.validate()

    ctx.obj['CONFIGURATION'] = config



from nalej_platformer.exceptions import AzureConfigPlanException, AzureProviderException
from nalej_platformer.utils import print_debug, spinner_ko_message, get_indentation, get_spinner
from nalej_platformer.provider import azure as azure_provider


class AzureConfigPlan:
    def __init__(self, resource_group, region, dns_zone, vm_size, aks_cluster_name=None):
        self.resource_group = resource_group
        self.region = region
        self.dns_zone = dns_zone
        self.vm_size = vm_size
        self.aks_cluster_name = aks_cluster_name

    def info(self, indentation=2):
        indent = get_indentation(indentation)

        info_template = (
            f"{indent}Resource Group: {self.resource_group}\n"
            f"{indent}Region: {self.region}\n"
            f"{indent}DNS Zone: {self.dns_zone}\n"
            f"{indent}Node VM Size: {self.vm_size}"

        )

        if self.aks_cluster_name is not None:
            info_template += f"\n{indent}AKS Cluster Name: {self.aks_cluster_name}."

        return info_template

    @classmethod
    def read_plan(cls, plan_data):
        print_debug(f"Reading azure configuration block. Raw data: {plan_data}")
        if plan_data is None:
            return None

        try:
            resource_group = cls._validate_resource_group(plan_data.get("resourceGroup", None))
            region = cls._validate_region(plan_data.get("region", None))
            dns_zone = cls._validate_dns_zone(plan_data.get("dnsZone", None))
            vm_size = cls._validate_vm_size(plan_data.get("vmSize", None), region)
            aks_cluster_name = cls._validate_aks_cluster_name(plan_data.get("aksName", None), resource_group)

        except AzureProviderException as ape:
            raise AzureConfigPlanException(f"Error checking with Azure: {ape}")

        return cls(resource_group, region, dns_zone, vm_size, aks_cluster_name)

    @classmethod
    def _validate_resource_group(cls, resource_group):
        with get_spinner("Checking resource group is valid...") as spinner:
            if resource_group in ["", None]:
                spinner_ko_message(spinner, "Resource group name is empty")
                raise AzureConfigPlanException("Resource group must not be empty")

            return azure_provider.check_resource_group(resource_group, spinner)

    @classmethod
    def _validate_region(cls, region):
        with get_spinner("Checking Azure region is valid...") as spinner:
            if region in ["", None]:
                spinner_ko_message(spinner, "Azure region name is empty")
                raise AzureConfigPlanException("Azure region must not be empty")

            return azure_provider.check_region(region, spinner)

    @classmethod
    def _validate_dns_zone(cls, dns_zone):
        with get_spinner("Checking DNS Zone is valid...") as spinner:
            if dns_zone in ["", None]:
                spinner_ko_message(spinner, "Azure DNS zone name is empty")
                raise AzureConfigPlanException("Azure DNS zone must not be empty")

            return azure_provider.check_dns_zone(dns_zone, spinner)

    @classmethod
    def _validate_vm_size(cls, vm_size, region):
        with get_spinner("Checking VM size is valid...") as spinner:
            if vm_size in ["", None]:
                spinner_ko_message(spinner, "Azure VM size name is empty")
                raise AzureConfigPlanException("Azure VM size name must not be empty")

            return azure_provider.check_vm_size(vm_size, region, spinner)

    @classmethod
    def _validate_aks_cluster_name(cls, cluster_name=None, resource_group=None):
        if cluster_name is None:
            return cluster_name

        with get_spinner("Checking AKS cluster name exists...") as spinner:
            if cluster_name in ["", None]:
                spinner_ko_message(spinner, "AKS cluster name is empty")
                raise AzureConfigPlanException("AKS cluster name must not be empty")

            return azure_provider.check_aks_cluster_name(cluster_name, resource_group, spinner)

    def dict_for_yaml(self):
        yaml_data = {
            "resourceGroup": self.resource_group,
            "region": self.region,
            "dnsZone": self.dns_zone,
            "vmSize": self.vm_size
        }

        if self.aks_cluster_name is not None:
            yaml_data["aksName"] = self.aks_cluster_name

        return yaml_data



import yaml
import click

from nalej_platformer.utils import print_debug, get_indentation
from nalej_platformer.plan.management import ManagementPlan
from nalej_platformer.plan.application import ApplicationPlan
from nalej_platformer.plan.organization import OrganizationPlan
from nalej_platformer.exceptions import PlanException


class Plan:
    def __init__(self, plan_data):
        self.management = ManagementPlan.read_plan(plan_data.get('management', None))
        self.organization_list = OrganizationPlan.read_list(plan_data.get('organization', None))
        self.application_list = ApplicationPlan.read_list(plan_data.get('application', None), self.organization_list)

    def info(self, indentation=0):
        indent = get_indentation(indentation)
        management = self.management.info(indentation=indentation + 1)
        info_template = f"{indent}Management:\n{management}"

        info_template += f"{indent}Application clusters:\n"
        for application in self.application_list:
            info_template += application.info(indentation=indentation + 1)

        info_template += f"{indent}Organizations:\n"
        for organization in self.organization_list:
            info_template += organization.info(indentation=indentation + 1)

        return info_template

    def apply(self):
        print_debug("Applying plan")

        self.management.apply()
        self.organization_list = OrganizationPlan.apply_list(self.organization_list, self.management)
        self.application_list = ApplicationPlan.apply_list(
            self.application_list, self.organization_list, self.management)

    @classmethod
    def read_plan(cls, plan_path):
        """Reads the plan file"""

        print_debug(f"Reading plan file: {plan_path}")

        plan = None
        with open(plan_path) as plan_file:
            try:
                plan_dict = yaml.load(plan_file, Loader=yaml.FullLoader)
            except yaml.YAMLError as yexc:
                raise PlanException(f"Error reading YAML plan file. Details: {yexc}")
            else:
                click.echo("Reading and validating plan")
                plan = cls(plan_dict)

        return plan



import time
from datetime import datetime, timedelta
import os
from tempfile import gettempdir

import click
import yaml

from nalej_platformer.plan.azure import AzureConfigPlan
from nalej_platformer.plan.organization import OrganizationPlan
from nalej_platformer.plan.k8s import KubernetesConfigPlan
from nalej_platformer.plan.ip_addresses import IPAddressesPlan
from nalej_platformer.exceptions import ApplicationPlanException, ProviderException
from nalej_platformer.utils import print_debug, spinner_ok_message, spinner_ko_message, get_indentation, get_spinner
from nalej_platformer.provider.public_api import login, provision_and_install, get_cluster_state
from nalej_platformer.provider.azure import create_service_principal, get_aks_credentials, get_cluster_by_cluster_id
from nalej_platformer.provider.azure import get_ingress_ip_from_resource_group
from nalej_platformer.provider.kubernetes import validate_cluster_installation, get_node_count


ENVIRONMENTS = [
    "PRODUCTION",
    "STAGING",
    "DEVELOPMENT"
]


class ApplicationPlan:
    def __init__(
            self, name, k8s_config, organization, cluster_id, azure_config=None, entrypoint=None, ip_addresses=None):
        self.name = name
        self.organization = organization
        self.cluster_id = cluster_id
        self.entrypoint = entrypoint
        self.azure_config = azure_config
        self.k8s_config = k8s_config
        self.ip_addresses = ip_addresses

        self._already_provisioned = False
        self._already_installed = False
        self._apply_data = {}

    def info(self, indentation=1):
        indent = get_indentation(indentation)
        org_name = self.organization.info(only_name=True)
        k8s_config = self.k8s_config.info(indentation=indentation + 1)
        info_template = (
            f"{indent}Name: {self.name}\n"
            f"{indent}Organization: {org_name}\n"
        )

        if self.cluster_id is not None:
            info_template += f'{indent}Cluster ID: {self.cluster_id}\n'

        if self.entrypoint is not None:
            info_template += f'{indent}Entrypoint: {self.entrypoint}\n'

        info_template += f"{indent}K8s configuration: \n{k8s_config}\n"

        if self.azure_config is not None:
            azure_config = self.azure_config.info(indentation=indentation + 1)
            info_template += f"{indent}Azure configuration: \n{azure_config}"

        if self.ip_addresses is not None:
            ip_addresses = self.ip_addresses.info(indentation=indentation + 1)
            info_template += f"{indent}IP Addresses: \n{ip_addresses}\n"

        info_template += '\n'
        return info_template

    def apply(self, organization_list, management):
        self._management = management
        login_address = self._management.hostnames.login
        environment = self._management.environment

        self._validate_organization(organization_list, self.organization.name)

        with get_spinner(f'Creating application cluster {self.name}...') as spinner:
            try:
                login(login_address, environment, self.organization.nalej_admin, spinner)
            except ProviderException as pexc:
                raise ApplicationPlanException(pexc)
            else:
                print_debug(f'Applying application block {self.name}')

                workflow = [
                    "_apply_create_service_principal",
                    "_apply_provision_and_install",
                    "_validate_provision_and_install",
                    "_set_application_info",
                    "_update_plan_file",
                    "_apply_validate_application_install",
                ]

                for step in workflow:
                    try:
                        getattr(self, step)()
                    except ProviderException as pexc:
                        raise ApplicationPlanException(f"Error applying plan. Details: {pexc}")

        return self

    def get_dns_zone(self):
        if self.azure_config is not None:
            return self.azure_config.dns_zone
        else:
            return self.dns_zone

    @classmethod
    def read_list(cls, plan_list=[], org_list=[]):
        app_list = []
        if plan_list is None:
            return app_list

        click.echo("APPLICATION CLUSTERS: ")
        for i, plan in enumerate(plan_list, start=1):
            click.echo(f"Application cluster {i}")
            app_list.append(cls.read_plan(plan, org_list))
            click.echo('\n')
        return app_list

    @classmethod
    def read_plan(cls, plan_data, org_list):
        print_debug(f"Reading application plan block. Raw data: {plan_data}")
        if plan_data is None:
            raise ApplicationPlanException("No block found in plan")
        name = cls._validate_name(plan_data.get("name", None))
        cluster_id = plan_data.get("clusterID", None)
        entrypoint = plan_data.get("entrypoint", None)
        azure_config = AzureConfigPlan.read_plan(plan_data.get("azure", None))
        region = azure_config.region if azure_config is not None else None
        k8s_config = KubernetesConfigPlan.read_plan(plan_data.get("kubernetes", None), region)
        ip_addresses = IPAddressesPlan.read_plan(plan_data.get("ipAddresses", None))
        organization_name = plan_data.get("organization", None)
        organization = cls._validate_organization(org_list, organization_name)

        return cls(name, k8s_config, organization, cluster_id, azure_config, entrypoint, ip_addresses)

    @classmethod
    def apply_list(cls, application_list, organization_list, management):
        print_debug("Applying application block")

        applied_list = [app.apply(organization_list, management) for app in application_list]
        # cls._update_plan_file(applied_list)
        return applied_list

    @classmethod
    def _validate_name(cls, name):
        with get_spinner("Checking application cluster name is valid...") as spinner:
            if name in ["", None]:
                spinner_ko_message(spinner, "Application cluster name is empty")
                raise ApplicationPlanException("Name must not be empty")

            spinner_ok_message(spinner, "Application cluster name is valid")
        return name

    @classmethod
    def _validate_organization(cls, org_list, name):
        if name is None:
            raise ApplicationPlanException("No organization is set for this application")
        return OrganizationPlan.find_by_name(org_list, name)

    def _apply_create_service_principal(self):
        with get_spinner("Getting a service principal for provisioning...") as spinner:
            self._apply_data["sp_path"] = create_service_principal(spinner)

    def _apply_provision_and_install(self):
        node_count = None
        with get_spinner(f'Provisioning and/or installing application cluster {self.name}...') as spinner:
            if self.k8s_config.kubeconfig is not None:
                spinner.text = "Checking if application cluster is already provisioned..."
                try:
                    node_count = get_node_count(self.k8s_config.kubeconfig)
                    cluster_state = get_cluster_state(
                        self._management.hostnames.api, self._management.environment, self.cluster_id, spinner)
                except ProviderException as pexc:
                    raise ApplicationPlanException(pexc)
                else:
                    if cluster_state == 'PROVISIONED':
                        self._already_provisioned = True
                        # TODO: Install a previously provisioned cluster
                        spinner_ok_message(spinner, "Application cluster is already provisioned")
                        raise ApplicationPlanException("Install only application cluster workflow not supported yet")
                    elif cluster_state == 'INSTALLED':
                        self._already_provisioned = True
                        if validate_cluster_installation(self.k8s_config.kubeconfig, spinner, True):
                            self._already_installed = True
            else:
                sp_path = self._apply_data.get("sp_path", "")
                self._apply_data["provision_result"] = provision_and_install(
                    sp_path, self.azure_config.vm_size, self.azure_config.region, self.name,
                    self.azure_config.resource_group, self.azure_config.dns_zone, self.k8s_config.version,
                    self.k8s_config.num_nodes, self._management.environment, self._management.hostnames.api,
                    self.organization.organization_id, spinner
                )

        try:
            if node_count is not None:
                self.k8s_config.validate_num_nodes(node_count)
                self._already_provisioned = True
                spinner_ok_message(spinner, "Already provisioned application cluster seems valid")
        except ProviderException as pexc:
            raise ApplicationPlanException(pexc)

    def _validate_provision_and_install(self):
        if not self._already_installed:
            start_time = datetime.now()
            with get_spinner('Validating application cluster provision and install...') as spinner:
                cluster_id = self._apply_data["provision_result"]["cluster_id"]
                for i in range(0, 36):
                    elapsed_time = datetime.now() - start_time
                    elapsed_time = elapsed_time - timedelta(microseconds=elapsed_time.microseconds)
                    spinner.text = f'Validating application cluster provision and install... [{elapsed_time}]'
                    try:
                        cluster_state = get_cluster_state(
                            self._management.hostnames.api, self._management.environment, cluster_id, spinner)
                    except ProviderException as pexc:
                        raise ApplicationPlanException(pexc)
                    else:
                        if cluster_state == "FAILURE":
                            spinner_ko_message(spinner, "Application cluster failed to provision and/or install")
                            raise ApplicationPlanException("Application cluster failed to provision and/or install")
                        elif cluster_state == "INSTALLED":
                            spinner_ok_message(spinner, "Application cluster finished provision and/or install")
                            return
                        else:
                            time.sleep(60)
                raise ApplicationPlanException("Timeout waiting for application cluster to provision and/or install")

    def _set_application_info(self):
        if not self._already_provisioned:
            provision_result = self._apply_data.get("provision_result", {})
            self.cluster_id = provision_result['cluster_id']

            with get_spinner('Getting cluster information from Azure...') as spinner:
                cluster_info = get_cluster_by_cluster_id(self.cluster_id, spinner)

            cluster_name = f"appcluster-{self.cluster_id}"
            self.azure_config.aks_cluster_name = cluster_name

            dns_zone = self.get_dns_zone()
            self.entrypoint = f'*.{self.name}.{dns_zone}'

            kubeconfig_path = os.path.join(gettempdir(), f'{cluster_name}.yaml')
            with get_spinner('Getting AKS cluster kubeconfig file...') as spinner:
                get_aks_credentials(self.azure_config.resource_group, cluster_name, kubeconfig_path, spinner)
                self.k8s_config._set_kubeconfig(kubeconfig_path, cluster_name)

            with get_spinner('Getting Ingress IP Address...') as spinner:
                ingress_address = get_ingress_ip_from_resource_group(cluster_info['nodeResourceGroup'], spinner)
                self.ip_addresses = IPAddressesPlan(ingress=ingress_address)

    def _update_plan_file(self):
        if not self._already_provisioned:
            with get_spinner("Updating plan file...") as spinner:
                ctx = click.get_current_context()
                yaml_contents = {}
                with open(ctx.obj['PLAN_PATH'], "r") as plan_file:
                    yaml_contents = yaml.load(plan_file, Loader=yaml.FullLoader)

                application_list = yaml_contents.get("application", None)
                if application_list is None:
                    spinner_ko_message(spinner, "The original plan file is malformed")
                    raise ApplicationPlanException("No application block found in plan file")

                for i, application_block in enumerate(application_list):
                    if application_block['name'] == self.name:
                        application_block['clusterID'] = self.cluster_id
                        application_block['entrypoint'] = self.entrypoint
                        application_block['ipAddresses'] = self.ip_addresses.dict_for_yaml()
                        application_block['kubernetes'] = self.k8s_config.dict_for_yaml()
                        application_block['azure'] = self.azure_config.dict_for_yaml()

                        application_list[i] = application_block

                yaml_contents["application"] = application_list

                with open(ctx.obj['PLAN_PATH'], "w") as plan_file:
                    plan_file.write(yaml.dump(yaml_contents, default_flow_style=False))

                spinner_ok_message(spinner, "Updated plan file")

    def _apply_validate_application_install(self):
        if not self._already_installed:
            with get_spinner("Validating cluster installation...") as spinner:
                validate_cluster_installation(self.k8s_config.kubeconfig, spinner)



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

with open('requirements.txt') as fp:
    install_requires = fp.read()

setuptools.setup(
    name="nalej-platformer",
    version="0.5.2",
    author="Rodrigo Núñez",
    author_email="rnunez@nalej.com",
    description="Nalej Platform automation",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/nalej/nalej-platformer",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
    install_requires=install_requires,
    python_requires='>=3.7',
    entry_points='''
        [console_scripts]
        nalej-platformer=nalej_platformer.cli.main:run
    '''
)



import yaml
import click

from nalej_platformer.plan.user import OrganizationUserPlan
from nalej_platformer.exceptions import OrganizationPlanException, SignupProviderException
from nalej_platformer.utils import print_debug, spinner_ko_message, spinner_ok_message, get_indentation, get_spinner
from nalej_platformer.provider.signup import wait_for_signup, create_organization, organization_exists


class OrganizationPlan:
    def __init__(self, name, owner, nalej_admin, organization_id=None):
        self.name = name
        self.owner = owner
        self.nalej_admin = nalej_admin
        self.organization_id = organization_id

    def info(self, only_name=False, indentation=2):
        if only_name:
            return self.name
        else:
            indent = get_indentation(indentation)
            owner = self.owner.info(indentation=indentation + 1)
            nalej_admin = self.nalej_admin.info(indentation=indentation + 1)
            info_template = (
                f"{indent}Name: {self.name}\n"
                f"{indent}Users:\n"
                f"{indent}- Owner:\n{owner}\n"
                f"{indent}- Nalej Admin\n{nalej_admin}"
            )
            if self.organization_id is not None:
                id_info = f"{indent}ID: {self.organization_id}\n"
                info_template = id_info + info_template

            return info_template

    def apply(self, management):
        signup_address = management.hostnames.signup
        environment = management.environment

        with get_spinner(f'Creating organization {self.name}...') as spinner:
            if organization_exists(signup_address, self, environment):
                spinner_ok_message(spinner, f"Organization {self.name} already exists, skipping")
            else:
                try:
                    self.organization_id = create_organization(signup_address, self, environment, spinner)
                except SignupProviderException as sexc:
                    raise OrganizationPlanException(sexc)

        return self

    @classmethod
    def read_list(cls, plan_list=[]):
        org_list = []
        if plan_list is None:
            return org_list

        click.echo("ORGANIZATIONS: ")
        for plan in plan_list:
            name = plan.get("name", "<NO NAME>")
            click.echo(f"Organization {name}")
            org_list.append(cls.read_plan(plan))
            click.echo('\n')
        return org_list

    @classmethod
    def read_plan(cls, plan_data):
        print_debug(f"Reading organization configuration block. Raw data: {plan_data}")
        if plan_data is None:
            raise OrganizationPlanException("Organization configuration block is required")

        name = cls._validate_name(plan_data.get("name", None))
        users = plan_data.get("users", None)
        organization_id = plan_data.get("organizationID", None)

        if users is None:
            raise OrganizationPlanException("Users block is required")

        owner = OrganizationUserPlan.read_plan(users.get("owner", None), "owner")
        nalej_admin = OrganizationUserPlan.read_plan(users.get("nalejAdmin", None), "nalej admin")

        return cls(name, owner, nalej_admin, organization_id)

    @classmethod
    def apply_list(cls, organization_list, management):
        print_debug("Applying organization block")
        with get_spinner("Waiting for signup to be ready...") as spinner:
            signup_address = management.hostnames.signup
            environment = management.environment
            try:
                wait_for_signup(signup_address, environment, spinner)
            except SignupProviderException as sexc:
                raise OrganizationPlanException(sexc)

        applied_list = [org.apply(management) for org in organization_list]
        cls._update_plan_file(applied_list)
        return applied_list

    @classmethod
    def _validate_name(cls, name):
        with get_spinner("Checking organization name...") as spinner:
            if name in ["", None]:
                spinner_ko_message(spinner, "Organization name is empty")
                raise OrganizationPlanException("Organization name must not be empty")

            spinner_ok_message(spinner, "Organization name is valid")
            return name

    @staticmethod
    def _update_plan_file(organization_list):
        with get_spinner("Updating plan file...") as spinner:
            ctx = click.get_current_context()
            yaml_contents = {}
            with open(ctx.obj['PLAN_PATH'], "r") as plan_file:
                yaml_contents = yaml.load(plan_file, Loader=yaml.FullLoader)

            yaml_contents["organization"] = [org.dict_for_yaml() for org in organization_list]

            with open(ctx.obj['PLAN_PATH'], "w") as plan_file:
                plan_file.write(yaml.dump(yaml_contents, default_flow_style=False))

            spinner_ok_message(spinner, "Updated plan file")

    @staticmethod
    def find_by_name(org_list, name):
        for org in org_list:
            if org.name == name:
                return org

        return None

    def dict_for_yaml(self):
        data_dict = {
            "name": self.name,
            "users": {
                "nalejAdmin": self.nalej_admin.dict_for_yaml(),
                "owner": self.owner.dict_for_yaml()
            }
        }
        if self.organization_id is not None:
            data_dict["organizationID"] = self.organization_id
        return data_dict



from nalej_platformer.exceptions import IPAddressesPlanException
from nalej_platformer.utils import print_debug, get_indentation


class IPAddressesPlan:
    def __init__(self, ingress, dns=None, coredns=None, vpn_server=None):
        self.ingress = ingress
        self.dns = dns
        self.coredns = coredns
        self.vpn_server = vpn_server

    def info(self, indentation=2):
        indent = get_indentation(indentation)
        info_template = f"{indent}Ingress: {self.ingress}\n"

        if self.dns is not None:
            info_template += f"{indent}DNS: {self.dns}\n"
        if self.coredns is not None:
            info_template += f"{indent}CoreDNS: {self.coredns}\n"
        if self.vpn_server is not None:
            info_template += f"{indent}VPN Server: {self.vpn_server}\n"

        return info_template

    @classmethod
    def read_plan(cls, plan_data):
        print_debug(f"Reading IP Address plan block. Raw data: {plan_data}")
        if plan_data is None:
            return None

        ingress = plan_data.get("ingress", None)
        dns = plan_data.get("dns", None)
        coredns = plan_data.get("coredns", None)
        vpn_server = plan_data.get("vpnServer", None)

        if ingress is None:
            raise IPAddressesPlanException("Ingress IP Address not found in plan")

        return cls(ingress, dns, coredns, vpn_server)

    def dict_for_yaml(self):
        yaml_data = {
            "ingress": self.ingress
        }

        if self.dns is not None:
            yaml_data["dns"] = self.dns
        if self.coredns is not None:
            yaml_data["coredns"] = self.coredns
        if self.vpn_server is not None:
            yaml_data["vpnServer"] = self.vpn_server

        return yaml_data



import os
import subprocess
import json
import getpass
import time

import click

from nalej_platformer.exceptions import AzureProviderException
from nalej_platformer.utils import print_debug, spinner_ko_message, spinner_ok_message, get_spinner
from nalej_platformer.config.defaults import SP_PATH


def _get_path():
    ctx = click.get_current_context()
    config = ctx.obj['CONFIGURATION']
    return config.az_path


def login():
    with get_spinner(text="Logging to Azure...") as spinner:
        username = os.environ.get("AZURE_USERNAME", None)
        password = os.environ.get("AZURE_PASSWORD", None)
        if username is None or password is None:
            spinner_ko_message(spinner, "Error logging in to Azure")
            raise AzureProviderException(
                "Azure username or password not defined. Please ensure that both environment variables are set")

        print_debug(f'Logging in to Azure using username: {username} and password: {"*"*len(password)}')
        try:
            cmd = f"{_get_path()} login --username {username} --password {password}"
            subprocess.run(cmd, shell=True, check=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        except subprocess.CalledProcessError as cpe:
            print_debug(f'Error logging in to Azure. Output: {cpe.stdout}. Error: {cpe.stderr}.')
            spinner_ko_message(spinner, "Error logging in to Azure")
            raise AzureProviderException("Error logging to Azure with the provided credentials")
        else:
            print_debug('Logged in to Azure successfully')
            spinner_ok_message(spinner, "Logged in to Azure")
            return True


def check_resource_group(resource_group, spinner):
    if resource_group in ["", None]:
        spinner_ko_message(spinner, "Error checking Azure Resource Group")
        raise AzureProviderException("Resource group name is empty")

    try:
        print_debug(f'Checking Azure Resource group: {resource_group} exists')
        cmd = f"{_get_path()} group show --name {resource_group}"
        subprocess.run(cmd, shell=True, check=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    except subprocess.CalledProcessError as cpe:
        print_debug(f'Error checking Azure Resource group. Output: {cpe.stdout}. Error: {cpe.stderr}.')
        spinner_ko_message(spinner, "Error checking Azure Resource Group")
        raise AzureProviderException("Error checking resource group")
    else:
        spinner_ok_message(spinner, "Azure Resource Group exists")
        return resource_group


def check_region(region, spinner):
    if region in ["", None]:
        spinner_ko_message(spinner, "Error checking Azure Region")
        raise AzureProviderException("Region name is empty")

    try:
        print_debug(f'Checking Azure Region {region} exists')
        cmd = (
            f"{_get_path()} account list-locations --query "
            f"\"[?displayName=='{region}' || name=='{region}'].displayName | [0]\""
        )
        get_region = subprocess.run(cmd, shell=True, check=True, capture_output=True, text=True)
    except subprocess.CalledProcessError as cpe:
        print_debug(f'Error checking Azure region. Output: {cpe.stdout}. Error: {cpe.stderr}.')
        spinner_ko_message(spinner, "Error checking Azure Region")
        raise AzureProviderException("Error checking Azure region")
    else:
        output = get_region.stdout.strip('"\n')
        if output == "":
            spinner_ko_message(spinner, "No matching region found")
            raise AzureProviderException(f'No matching region found for {region}')
        spinner_ok_message(spinner, f"Found matching Azure region: {output}")
        return output


def check_dns_zone(dns_zone, spinner):
    if dns_zone in ["", None]:
        spinner_ko_message(spinner, "Error checking DNS Zone")
        raise AzureProviderException("DNS Zone is empty")

    try:
        print_debug(f'Checking Azure DNS Zone {dns_zone} exists')
        cmd = f"{_get_path()} network dns zone list --query \"[?name=='{dns_zone}'] | [0].[name, resourceGroup]\""
        get_dns_zone = subprocess.run(cmd, shell=True, check=True, capture_output=True, text=True)
    except subprocess.CalledProcessError as cpe:
        print_debug(f'Error checking DNS Zone. Output: {cpe.stdout}. Error: {cpe.stderr}.')
        spinner_ko_message(spinner, "Error checking DNS Zone")
        raise AzureProviderException("Error checking DNS Zone")
    else:
        output = get_dns_zone.stdout.strip('"\n')
        if output == "":
            spinner_ko_message(spinner, "No matching DNS Zone found")
            raise AzureProviderException(f'No matching region found for {dns_zone}')

        output_json = json.loads(output)
        matched_zone = output_json[0]
        matched_resource_group = output_json[1]

        spinner_ok_message(spinner, f"Found DNS Zone: {matched_zone} in {matched_resource_group} resource group")
        return matched_zone


def check_vm_size(vm_size, region, spinner):
    if vm_size in ["", None]:
        spinner_ko_message(spinner, "Error checking VM size")
        raise AzureProviderException("VM size is empty")
    if region in ["", None]:
        spinner_ko_message(spinner, "Error checking VM size")
        raise AzureProviderException("Region name is empty")

    try:
        print_debug(f'Checking VM Size {vm_size} for {region} Azure Region')
        cmd = f"{_get_path()} vm list-sizes --location '{region}' --query \"[?name=='{vm_size}'].name | [0]\""
        get_vm_size = subprocess.run(cmd, shell=True, check=True, capture_output=True, text=True)
    except subprocess.CalledProcessError as cpe:
        print_debug(f'Error checking VM size. Output: {cpe.stdout}. Error: {cpe.stderr}.')
        spinner_ko_message(spinner, "Error checking VM size")
        raise AzureProviderException("Error checking VM size")
    else:
        output = get_vm_size.stdout.strip('"\n')
        if output == "":
            spinner_ko_message(spinner, "No matching VM size found")
            raise AzureProviderException(f'No matching VM size found for {vm_size} in {region} region')

        spinner_ok_message(spinner, f"Found VM Size: {output} in {region} region")
        return output


def check_k8s_version(version, region, spinner):
    if version in ["", None]:
        spinner_ko_message(spinner, "Error checking Kubernetes version")
        raise AzureProviderException("Kubernetes version is empty")
    if region in ["", None]:
        spinner_ko_message(spinner, "Error checking Kubernetes version")
        raise AzureProviderException("Region name is empty")

    try:
        print_debug(f'Checking K8s version {version} is available in {region} Azure Region')
        cmd = (
            f"{_get_path()} aks get-versions --location '{region}' --query "
            f"\"orchestrators[?orchestratorType=='Kubernetes' && orchestratorVersion=='{version}'] | [0]\""
        )
        get_k8s_version = subprocess.run(cmd, shell=True, check=True, capture_output=True, text=True)
    except subprocess.CalledProcessError as cpe:
        print_debug(f'Error checking Kubernetes version. Output: {cpe.stdout}. Error: {cpe.stderr}.')
        spinner_ko_message(spinner, "Error checking Kubernetes version")
        raise AzureProviderException("Error checking Kubernetes version")
    else:
        output = get_k8s_version.stdout.strip('"\n')
        if output == "":
            spinner_ko_message(spinner, "No matching Kubernetes version found")
            raise AzureProviderException(f'No matching region found for {version} in {region} region')

        output_json = json.loads(output)
        matched_version = output_json["orchestratorVersion"]
        is_preview = output_json["isPreview"]
        if is_preview:
            spinner_ko_message(spinner, "Kubernetes version is not stable")
            raise AzureProviderException(f'Kubernetes version {version} is in preview state and not production ready')

        spinner_ok_message(spinner, f"Found Kubernetes version: {matched_version} in {region} region")
        return matched_version


def check_aks_cluster_name(cluster_name, resource_group, spinner, required=True):
    if cluster_name in ["", None]:
        spinner_ko_message(spinner, "Error checking AKS cluster name")
        raise AzureProviderException("Cluster name is empty")
    if resource_group in ["", None]:
        spinner_ko_message(spinner, "Error checking AKS cluster name")
        raise AzureProviderException("Resource group is empty")

    try:
        print_debug(f'Checking if AKS service {cluster_name} exists in {resource_group} Azure Resource Group')
        cmd = f"{_get_path()} aks show --resource-group {resource_group} --name {cluster_name}"
        subprocess.run(cmd, shell=True, check=True, capture_output=True, text=True)
    except subprocess.CalledProcessError as cpe:
        if required:
            print_debug(f'Error getting AKS cluster name. Output: {cpe.stdout}. Error: {cpe.stderr}.')
            spinner_ko_message(spinner, "Error getting AKS cluster name")
            raise AzureProviderException("Error getting AKS cluster name")
        else:
            spinner_ok_message(spinner, "AKS cluster name does not exist")
    else:
        spinner_ok_message(spinner, f"AKS cluster name {cluster_name} exists")

    return cluster_name


def create_service_principal(spinner):
    if os.path.exists(SP_PATH):
        print_debug("Service Principal for provisioning already exists")
        spinner_ok_message(spinner, "Service Principal already exists")
    else:
        print_debug("No Service Principal for provisioning found, creating")
        sp_name = f"{getpass.getuser()}_provisioner"

        try:
            print_debug(f'Creating new Service Principal with name {sp_name}')
            cmd = f"{_get_path()} ad sp create-for-rbac --name http://{sp_name} --sdk-auth"
            create_sp = subprocess.run(cmd, shell=True, check=True, capture_output=True, text=True)
        except subprocess.CalledProcessError as cpe:
            print_debug(f'Error creating provisioner service principal. Output: {cpe.stdout}. Error: {cpe.stderr}.')
            spinner_ko_message(spinner, "Error creating Azure Service Principal")
            raise AzureProviderException("Error creating Azure Service Principal")
        else:
            with open(SP_PATH, "w") as sp_file:
                sp_file.write(create_sp.stdout)
                print_debug(f'Provisioner Service Principal file created at {SP_PATH}')

            time.sleep(30)  # Sleep to give enough time for Azure to really create the SP
        spinner_ok_message(spinner, "Service Principal created successfully")
    return SP_PATH


def get_cluster_by_cluster_id(cluster_id, spinner):
    try:
        print_debug(f'Getting AKS cluster information for Cluster ID: {cluster_id}')

        cmd = f'{_get_path()} aks list --query "[?tags.clusterID==\'{cluster_id}\']"'
        cluster_info = subprocess.run(cmd, shell=True, check=True, capture_output=True, text=True)
    except subprocess.CalledProcessError as cpe:
        print_debug(f'Error getting cluster info from Azure. Output: {cpe.stdout}. Error: {cpe.stderr}.')
        spinner_ko_message(spinner, "Error getting cluster info from Azure")
        raise AzureProviderException("Error getting cluster info from Azure")
    else:
        try:
            json_info = json.loads(cluster_info.stdout)
        except json.JSONDecodeError:
            raise AzureProviderException("Malformed response getting cluster info from Azure")
        else:
            spinner_ok_message(spinner, "Retrieved cluster information from Azure successfully")
            return json_info[0]


def get_aks_credentials(resource_group, cluster_name, kubeconfig_path, spinner):
    try:
        print_debug(f'Getting AKS cluster kubeconfig file cluster: {cluster_name} on {resource_group} resource group')

        cmd = (
            f'{_get_path()} aks get-credentials --resource-group {resource_group} '
            f'--name {cluster_name} --file {kubeconfig_path} --overwrite-existing'
        )
        subprocess.run(cmd, shell=True, check=True, capture_output=True, text=True)
    except subprocess.CalledProcessError as cpe:
        print_debug(f'Error getting AKS cluster kubeconfig file. Output: {cpe.stdout}. Error: {cpe.stderr}.')
        spinner_ko_message(spinner, "Error getting AKS cluster kubeconfig file")
        raise AzureProviderException("Error getting AKS cluster kubeconfig file")
    else:
        spinner_ok_message(spinner, "Retrieved AKS cluster kubeconfig file successfully")


def get_ingress_ip_from_resource_group(resource_group, spinner):
    try:
        print_debug(f'Getting Ingress IP Address from {resource_group} resource group')

        cmd = f'{_get_path()} network public-ip show --resource-group {resource_group} --name ingressPublicIPAddress'
        ip_info = json.loads(subprocess.run(cmd, shell=True, check=True, capture_output=True, text=True).stdout)
    except subprocess.CalledProcessError as cpe:
        print_debug((
            f'Error getting Ingress IP Address from {resource_group} resource group. '
            f'Output: {cpe.stdout}. Error: {cpe.stderr}.'
        ))
        spinner_ko_message(spinner, f'Error getting Ingress IP Address from {resource_group} resource group')
        raise AzureProviderException(f'Error getting Ingress IP Address from {resource_group} resource group')
    except json.JSONDecodeError:
        print_debug(f'Malformed response getting ingress IP Address info.')
        spinner_ko_message(spinner, "Malformed response getting ingress IP Address info")
        raise AzureProviderException("Malformed response getting ingress IP Address info")
    else:
        spinner_ok_message(spinner, "Found Ingress IP Address")
        return ip_info['ipAddress']



import click
import yaml

from nalej_platformer.plan.azure import AzureConfigPlan
from nalej_platformer.plan.k8s import KubernetesConfigPlan
from nalej_platformer.plan.hostnames import HostnamesPlan
from nalej_platformer.plan.ip_addresses import IPAddressesPlan
from nalej_platformer.exceptions import ManagementPlanException, ProviderException
from nalej_platformer.utils import rreplace, print_debug, spinner_ok_message, spinner_ko_message, get_indentation
from nalej_platformer.utils import get_spinner
from nalej_platformer.provider.azure import create_service_principal
from nalej_platformer.provider.provisioner import provision_management_cluster
from nalej_platformer.provider.installer import install_management_cluster
from nalej_platformer.provider.kubernetes import validate_cluster_installation, get_node_count


ENVIRONMENTS = [
    "PRODUCTION",
    "STAGING",
    "DEVELOPMENT"
]


class ManagementPlan:
    def __init__(
            self, name, k8s_config, environment, azure_config=None, hostnames=None, ip_addresses=None):
        self.name = name
        self.azure_config = azure_config
        self.k8s_config = k8s_config
        self.environment = environment
        self.hostnames = hostnames
        self.ip_addresses = ip_addresses

        self._already_provisioned = False
        self._already_installed = False
        self._apply_data = {}

    def info(self, indentation=1):
        indent = get_indentation(indentation)
        k8s_config = self.k8s_config.info(indentation=indentation + 1)

        info_template = (
            f"{indent}Name: {self.name}\n"
            f"{indent}Environment: {self.environment}\n"
            f"{indent}K8s configuration: \n{k8s_config}\n"
        )

        if self.azure_config is not None:
            azure_config = self.azure_config.info(indentation=indentation + 1)
            info_template += f"{indent}Azure configuration: \n{azure_config}\n"

        if self.hostnames is not None:
            hostnames = self.hostnames.info(indentation=indentation + 1)
            info_template += f"{indent}Hostnames: \n{hostnames}"

        if self.ip_addresses is not None:
            ip_addresses = self.ip_addresses.info(indentation=indentation + 1)
            info_template += f"{indent}IP Addresses: \n{ip_addresses}\n"

        return info_template

    def apply(self):
        workflow = [
            "_apply_create_service_principal",
            "_apply_provision_management_cluster",
            "_set_management_info",
            "_update_plan_file",
            "_apply_install_management_cluster",
            "_apply_validate_management_install",
        ]
        print_debug("Applying management block")

        for step in workflow:
            try:
                getattr(self, step)()
            except ProviderException as pexc:
                raise ManagementPlanException(f"Error applying plan. Details: {pexc}")

    @classmethod
    def read_plan(cls, plan_data):
        print_debug(f"Reading management plan block. Raw data: {plan_data}")
        click.echo("MANAGEMENT CLUSTER: ")
        if plan_data is None:
            raise ManagementPlanException("No block found in plan")

        name = cls._validate_name(plan_data.get("name", None))
        environment = cls._validate_environment(plan_data.get("environment", None))
        azure_config = AzureConfigPlan.read_plan(plan_data.get("azure", None))
        region = None if azure_config is None else azure_config.region
        k8s_config = KubernetesConfigPlan.read_plan(plan_data.get("kubernetes", None), region)
        hostnames = HostnamesPlan.read_plan(plan_data.get("hostnames", None))
        ip_addresses = IPAddressesPlan.read_plan(plan_data.get("ipAddresses", None))
        dns_zone = cls._validate_dns_zone(plan_data.get("dns_zone", None))

        if azure_config is None and dns_zone is None:
            raise ManagementPlanException("No DNS Zone defined or Azure block containing it found")
        if azure_config is not None and dns_zone is not None:
            raise ManagementPlanException("Duplicated DNS Zone defined in both Azure block and DNS Zone field")

        click.echo('\n')

        return cls(name, k8s_config, environment, azure_config, hostnames, ip_addresses)

    def get_dns_zone(self):
        if self.azure_config is not None:
            return self.azure_config.dns_zone
        else:
            return self.dns_zone

    @staticmethod
    def _validate_name(name):
        with get_spinner("Checking management cluster name is valid...") as spinner:
            if name in ["", None]:
                spinner_ko_message(spinner, "Management cluster name is empty")
                raise ManagementPlanException("Name must not be empty")

            spinner_ok_message(spinner, "Management cluster name is valid")
        return name

    @staticmethod
    def _validate_dns_zone(dns_zone):
        if dns_zone is not None:
            with get_spinner("Checking management cluster DNS Zone is valid...") as spinner:
                if dns_zone in ["", None]:
                    spinner_ko_message(spinner, "Management cluster DNS Zone is empty")
                    raise ManagementPlanException("Management cluster DNS Zone must not be empty")

                spinner_ok_message(spinner, "Management cluster DNS Zone is valid")
        return dns_zone

    @staticmethod
    def _validate_environment(environment):
        with get_spinner("Checking platform environment is valid...") as spinner:
            if environment in ["", None]:
                spinner_ko_message(spinner, "Environment is empty")
                raise ManagementPlanException("Environment must not be empty")

            environment = environment.upper()
            if environment not in ENVIRONMENTS:
                env_list = rreplace(', '.join(ENVIRONMENTS), ', ', ' or ', 1)
                spinner_ko_message(spinner, "Environment is not valid")
                raise ManagementPlanException(f"Invalid environment. Must be {env_list}")

            spinner_ok_message(spinner, f"Environment: {environment}")
        return environment

    def _apply_create_service_principal(self):
        with get_spinner("Getting a service principal for provisioning...") as spinner:
            self._apply_data["sp_path"] = create_service_principal(spinner)

    def _apply_provision_management_cluster(self):
        node_count = None
        with get_spinner("Provisioning management cluster...") as spinner:
            if self.k8s_config.kubeconfig is not None:
                spinner.text = "Checking if management cluster is already provisioned..."
                try:
                    node_count = get_node_count(self.k8s_config.kubeconfig)
                except ProviderException as pexc:
                    raise ManagementPlanException(pexc)
                else:
                    spinner_ok_message(spinner, "Management cluster is already provisioned")
            else:
                sp_path = self._apply_data.get("sp_path", "")
                try:
                    self._apply_data["provision_result"] = provision_management_cluster(
                        sp_path, self.azure_config.vm_size, self.azure_config.region, self.name,
                        self.azure_config.resource_group, self.azure_config.dns_zone, self.k8s_config.version,
                        self.k8s_config.num_nodes, self.environment, spinner
                    )
                except ProviderException as pexc:
                    raise ManagementPlanException(pexc)

        try:
            if node_count is not None:
                self.k8s_config.validate_num_nodes(node_count)
                self._already_provisioned = True
                spinner_ok_message(spinner, "Already provisioned management cluster seems valid")
        except ProviderException as pexc:
            raise ManagementPlanException(pexc)

    def _set_management_info(self):
        if not self._already_provisioned:
            provision_result = self._apply_data.get("provision_result", {})
            dns_zone = self.get_dns_zone()
            cluster_name = f"mngt-{self.name}"

            self.k8s_config._set_kubeconfig(provision_result.get("kubeconfig", None), cluster_name)
            self.azure_config.aks_cluster_name = cluster_name

            self.hostnames = HostnamesPlan(**{
                "web": f"web.{self.name}.{dns_zone}",
                "api": f"api.{self.name}.{dns_zone}",
                "login": f"login.{self.name}.{dns_zone}",
                "signup": f"signup.{self.name}.{dns_zone}"
            })

            self.ip_addresses = IPAddressesPlan(**{
                "ingress": provision_result.get("ingress_ip", None),
                "dns": provision_result.get("dns_ip", None),
                "coredns": provision_result.get("coredns_ip", None),
                "vpn_server": provision_result.get("vpnserver_ip", None)
            })

    def _update_plan_file(self):
        if not self._already_provisioned:
            with get_spinner("Updating plan file...") as spinner:
                ctx = click.get_current_context()
                yaml_contents = {}
                with open(ctx.obj['PLAN_PATH'], "r") as plan_file:
                    yaml_contents = yaml.load(plan_file, Loader=yaml.FullLoader)

                management_block = yaml_contents.get("management", None)
                if management_block is None:
                    spinner_ko_message(spinner, "The original plan file is malformed")
                    raise ManagementPlanException("No management block found in plan file")
                management_block["hostnames"] = self.hostnames.dict_for_yaml()
                management_block["ipAddresses"] = self.ip_addresses.dict_for_yaml()
                management_block["kubernetes"] = self.k8s_config.dict_for_yaml()
                management_block["azure"] = self.azure_config.dict_for_yaml()

                yaml_contents["management"] = management_block

                with open(ctx.obj['PLAN_PATH'], "w") as plan_file:
                    plan_file.write(yaml.dump(yaml_contents, default_flow_style=False))

                spinner_ok_message(spinner, "Updated plan file")

    def _apply_install_management_cluster(self):
        with get_spinner("Installing management cluster...") as spinner:
            if self._already_provisioned:
                spinner.text = "Checking if Nalej is installed and running"
                if validate_cluster_installation(self.k8s_config.kubeconfig, spinner):
                    self._already_installed = True
                    return

            install_management_cluster(
                self.k8s_config.kubeconfig, self.name, self.environment, self.get_dns_zone(),
                self.ip_addresses.ingress, self.ip_addresses.dns, self.ip_addresses.coredns,
                self.ip_addresses.vpn_server, spinner)

    def _apply_validate_management_install(self):
        if not self._already_installed:
            with get_spinner("Validating cluster installation...") as spinner:
                validate_cluster_installation(self.k8s_config.kubeconfig, spinner)



import subprocess
from datetime import datetime, timedelta

import click

from nalej_platformer.utils import spinner_ok_message, spinner_ko_message, print_debug


def _get_path():
    ctx = click.get_current_context()
    config = ctx.obj['CONFIGURATION']
    return config.installer_path


def _get_assets_path():
    ctx = click.get_current_context()
    config = ctx.obj['CONFIGURATION']
    return config.installer_resources


def _get_binaries_path():
    ctx = click.get_current_context()
    config = ctx.obj['CONFIGURATION']
    return config.installer_binaries

def _get_istio_path():
    return f'{_get_binaries_path()}/istio/bin'


def install_management_cluster(
        kubeconfig, name, environment, dns_zone, ingress_ip, dns_ip, coredns_ip, vpnserver_ip, spinner):

    start_time = datetime.now()
    cmd = (
        f'{_get_path()} install management --debug --binaryPath {_get_binaries_path()} '
        f'--componentsPath {_get_assets_path()} --managementClusterPublicHost={name}.{dns_zone} '
        f'--dnsClusterPublicHost=dns.{name}.{dns_zone} --targetPlatform=AZURE --useStaticIPAddresses '
        f'--ipAddressIngress={ingress_ip} --ipAddressDNS={dns_ip} --ipAddressCoreDNS={coredns_ip} '
        f'--ipAddressVPNServer={vpnserver_ip} --kubeConfigPath={kubeconfig} --targetEnvironment={environment} '
        f'--networkingMode=istio --istioPath={_get_istio_path()}'
    )

    install_mngt = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True, shell=True)

    result = False
    while True:
        if install_mngt.poll() is not None:
            break

        stderr_line = install_mngt.stderr.readline()
        debug_line = stderr_line.strip('\n')
        if debug_line != "":
            print_debug(f"Installer output: {debug_line}")
        elapsed_time = datetime.now() - start_time
        elapsed_time = elapsed_time - timedelta(microseconds=elapsed_time.microseconds)
        spinner.text = f'Installing management cluster...[{elapsed_time}]'

    install_mngt.stdout.close()
    install_mngt.stderr.close()
    exit_code = install_mngt.poll()
    if exit_code == 0:
        spinner_ok_message(spinner, "Management cluster installed successfully")
        result = True
    else:
        spinner_ko_message(spinner, "Error installing management cluster")

    return result
